#include <sys/fcntl.h>
#include <sys/stat.h>

#ifdef WIN32
# include <winsock2.h>
# include <process.h>
# include <windows.h>
#else
# include <sys/poll.h>
# include <sys/socket.h>
# include <netdb.h>
# include <netinet/in.h>
# include <arpa/inet.h> /* for inet_ntoa */
# include <sys/wait.h>
# include <sys/socket.h>
#endif

#include <signal.h>
#include <unistd.h>
#include <stdlib.h>

#include <vector>
#include <cerrno>
#include <cstdio>
#include <string.h> /* for memset */
#include <set>
#include <utility>

#include "config.hh"
#include "cset.hh"

/* Configuration */

struct sockaddr *ResolveHost(const char *name, int family)
{
    struct addrinfo *ai = NULL, hints;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = family;
    hints.ai_flags = AI_ADDRCONFIG;

    if (!getaddrinfo(name, NULL, &hints, &ai) && ai)
    {
        void *tmp = malloc(ai->ai_addrlen);
        memcpy(tmp, ai->ai_addr, ai->ai_addrlen);

        freeaddrinfo(ai);

        return (struct sockaddr*) tmp;
    }
    return NULL;
}

static const Chain2& FindUTF8()
{
    /* This must always work. If this throws exceptions, we're doomed. */
    static const Chain2 utf8 = CreateBuiltinChainPair("utf-8");
    return utf8;
}

static const Chain2& FindDefaultChain(const std::string& defaultname)
{
    static std::map<std::string, Chain2> cache;

    std::map<std::string, Chain2>::const_iterator i;
    i = cache.find(defaultname);

    if(i != cache.end())
    {
        return i->second;
    }

    try
    {
        const Chain2& Default = FindChain2(defaultname);
        return cache[defaultname] = Default;
    }
    catch(const std::string& chainname)
    {
        fprintf(stderr, "Default chain '%s' not defined! Trying internal chains\n",
            defaultname.c_str());

        try
        {
            const Chain2& Internal = CreateBuiltinChainPair(defaultname);
            fprintf(stderr, "Found internal chain.\n");
            return cache[defaultname] = Internal;
        }
        catch(const std::string& i_chainname)
        {
            fprintf(stderr, "No internal chain '%s'! Falling back to internal default 'UTF-8'\n",
                defaultname.c_str());
            return cache[defaultname] = FindUTF8();
        }
    }
}

static class ChannelCache
{
    std::map<std::string, Chain2> cache;
public:
    ChannelCache(): cache()
    {
        Reload();
    }
    void Reload()
    {
        cache.clear();

        ConfParser::FieldType specials = GetConf("settings", "special_chain");
        for(unsigned a=0; a<specials.size(); ++a)
        {
            const ConfParser::FuncType& func = specials[a];

            if(func.size() <= 1)
            {
                continue;
            }

            try
            {
                Chain2 chain2 = FindChain2(func[0].SField);
                for(unsigned b=1; b<func.size(); ++b)
                {
                    cache[func[b].SField] = chain2;
                }
            }
            catch(const std::string& chainname)
            {
                fprintf(stderr, "Chain '%s' is not defined, although referenced in special_chain settings.\n",
                    chainname.c_str());
            }
        }
    }
    const Chain2& FindForChannel(const std::string& channame) const
    {
        std::map<std::string, Chain2>::const_iterator i;
        i = cache.find(channame);

        if(i != cache.end()) return i->second;

        static const std::string defaultname = GetConf("settings", "default_chain").SField();
        return FindDefaultChain(defaultname);
    }

    const std::string Explain() const
    {
        std::string result;
        for(std::map<std::string,Chain2>::const_iterator
            i = cache.begin(); i != cache.end(); ++i)
        {
            if(!result.empty()) result += "; ";
            result += i->first + ":" + i->second.name;
        }
        if(result.empty())
        {
            result += "No special pairs defined";
        }
        return result;
    }
} ChannelCache;

void ReloadChannels()
{
    ChannelCache.Reload();
}

static const Chain2& FindChain2For(const std::string& channame)
{
    return ChannelCache.FindForChannel(channame);
}

static const std::string GetServer()
{
    return GetConf("settings", "server_host").SField();
}
static unsigned GetServerPort()
{
    return GetConf("settings", "server_port").IField();
}
static unsigned GetListenPort()
{
    return GetConf("settings", "listen_port").IField();
}

/* Util */

static const std::string AsciiUcase(const std::string& text)
{
    std::string result=text;
    for(unsigned a=0; a<result.size(); ++a)
    {
        char ch = result[a];
        if(ch >= 'a' && ch <= 'z') result[a] = ch+'A'-'a';
    }
    return result;
}

static const std::vector<std::string> Split(const std::string& text, char separator)
{
    std::vector<std::string> words;
    std::string::size_type beginpos = 0;
    while(beginpos < text.size())
    {
        std::string::size_type spacepos = text.find(separator, beginpos);
        if(spacepos == text.npos)break;

        words.push_back(text.substr(beginpos, spacepos-beginpos));

        beginpos = spacepos+1;
    }
    if(beginpos < text.size())
    {
        std::string rest = text.substr(beginpos);
        std::string::size_type endpos = rest.find("\r\n");
        if(endpos != rest.npos) rest.erase(endpos, rest.size());
        else {
            endpos = rest.find('\n');
            if(endpos != rest.npos) rest.erase(endpos, rest.size());
        }
        words.push_back(rest);
    }
    return words;
}


/* From client to server */
static const std::string OutChannel(const std::string& chan, const std::string& line)
{
    static const std::string client_enc = GetConf("settings", "client_chain").SField();

    const Chain2& server = FindChain2For(chan);
    const Chain2& client = FindDefaultChain(client_enc);

    return server.out.Encode(client.in.Decode(line));
}


/* From server to client */
static const std::string InChannel(const std::string& chan, const std::string& line)
{
    static const std::string client_enc = GetConf("settings", "client_chain").SField();

    const Chain2& server = FindChain2For(chan);
    const Chain2& client = FindDefaultChain(client_enc);

    return client.out.Encode(server.in.Decode(line));
}

static void ConvertMessage(const std::string& line,
    std::string& result, //! Message to be sent to recipient
    std::string& bounce  //! Message to be sent back to the sender
     )
{
    std::string guess_channel;
    std::string guess_command;

    enum { in, out } guess_dir = out;

    std::vector<std::string> words = Split(line, ' ');

    for(unsigned a=0; a<4; ++a)
    {
        if(a >= words.size()) break;
        const std::string& word = words[a];
        if(word.empty()) continue;

        if(a==0) guess_command = AsciiUcase(word);

        char ch = word[0];
        if(a == 0 && ch == ':') /* If the line begins with ':', it's from server probably */
        {
            guess_dir = in; // from server
        }
        if(ch == '#' || ch == '&' || ch == '+' || ch == '!'
        || (a==1 && guess_dir == out))
        {
            if(ch == '!' && word.size() > 6
               )
                guess_channel = "!" + word.substr(6);
            else
                guess_channel = word;
            break;
        }
    }

    if(guess_dir == out)
    {
        static const std::string me = "iconvircproxy";
        static const std::string you = "you";
        if(guess_channel == me)
        {
            std::string::size_type colonpos = line.find(':');
            std::string guess_message;
            if(colonpos != line.npos)
                guess_message = line.substr(colonpos+1, line.size()-colonpos-2); // eat '\n'
            words = Split(guess_message, ' ');

            std::string preamble = ":"+me+" ";
            if(guess_command == "WHOIS")
            {
                bounce = preamble+"311 "+you+" "+me+" * * :iconvircproxy "VERSION"\n"
                       + preamble+"319 "+you+" "+me+" :\n"
                       + preamble+"312 "+you+" "+me+" "+me+" :"+me+"\n"
                       + preamble+"318 "+you+" "+me+" :End of WHOIS list.\n";
                return;
            }
            if(guess_command == "PRIVMSG")
            {
                std::string command = words.size()>0 ? AsciiUcase(words[0]) : "";
                if(command == "RELOAD")
                {
                    ReloadConfig();
                    ReloadChainSetup();
                    ReloadChannels();
                    bounce = preamble+"NOTICE "+you+" :Configuration reloaded\n";
                }
                else if(command == "SHOW")
                {
                    if(words.size() <= 1)
                        bounce += preamble+"NOTICE "+you+" :"+ChannelCache.Explain()+"\n";

                    for(unsigned a=1; a<words.size(); ++a)
                    {
                        const Chain2& pair = FindChain2For(words[a]);
                        bounce += preamble+"NOTICE "+you+" :"+words[a]+": Pair "+pair.name+"\n";
                    }
                }
                else if(command == "SHOWPAIR")
                {
                    for(unsigned a=1; a<words.size(); ++a)
                    {
                        const Chain2& pair = FindDefaultChain(words[a]);

                        std::string in_str = pair.in.Explain();
                        std::string out_str = pair.out.Explain();

                        bounce += preamble+"NOTICE "+you+" :"+words[a]+": In "+in_str+", Out "+out_str+"\n";
                    }
                }
                else
                {
                    bounce = preamble+"NOTICE "+you+" :Unknown command '"+command+"'\n";
                    bounce+= preamble+"NOTICE "+you+" :RELOAD\n";
                    bounce+= preamble+"NOTICE "+you+" :SHOW <target>\n";
                    bounce+= preamble+"NOTICE "+you+" :SHOWPAIR <pairname>\n";
                }
                return;
            }
        }
    }

/*
    fprintf(stderr, "guess(chan=%s, dir=%s)\n",
        guess_channel.c_str(),
        guess_dir==in ? "in" : "out");
*/
    result =
        guess_dir == in ?  InChannel(guess_channel, line)
                        : OutChannel(guess_channel, line);
}

class SockStream
{
#ifndef WIN32
    mutable FILE* fp;
#else
    mutable int fd;
#endif
public:
    SockStream()
#ifndef WIN32
      : fp(NULL)
#else
      : fd(-1)
#endif
    {
    }
    SockStream(int sock)
#ifndef WIN32
      : fp(NULL)
#else
      : fd(-1)
#endif
    {
#ifndef WIN32
        // Nonblocking is not wanted
        fcntl(sock, F_SETFL,  fcntl(sock, F_GETFL) & ~O_NONBLOCK);
        fp = fdopen(sock, "r+");
        // Buffering is not wanted
        setbuf(fp,  NULL);
#else
        fd=sock;
#endif
    }
    SockStream(const SockStream& b)
#ifndef WIN32
      : fp(NULL)
#else
      : fd(-1)
#endif
    {
#ifndef WIN32
        fp=b.fp; b.fp=NULL;
#else
        fd=b.fd; b.fd=-1;
#endif
    }
    ~SockStream()
    {
#ifndef WIN32
        if(fp)fclose(fp);
#else
        if(fd>=0)closesocket(fd);
#endif
    }
    SockStream& operator =(const SockStream& b)
    {
#ifndef WIN32
        fp=b.fp; b.fp=NULL;
#else
        fd=b.fd; b.fd=-1;
#endif
        return *this;
    }
    const std::string ReadLine()
    {
#ifndef WIN32
        char Buf[8192];
        if(!fgets(Buf, sizeof Buf, fp)) return "";
        return Buf;
#else
        std::string result;
        for(;;)
        {
            char byte;
            int i = recv(fd, &byte, 1, 0);
            if(i == 1) { result += byte; if(byte == '\n') break; continue; }
            break;
        }
        return result;
#endif
    }
    void Write(const std::string& Buf)
    {
#ifndef WIN32
        fwrite(Buf.c_str(), 1, Buf.size(), fp);
#else
        send(fd, Buf.c_str(), Buf.size(), 0);
#endif
    }
    int GetFD() const
    {
#ifndef WIN32
        return fileno(fp);
#else
        return fd;
#endif
    }
};

static bool ReadWrite(SockStream& inf, SockStream& outf)
{
    std::string line = inf.ReadLine();
    if(line.empty()) return true;

    std::string::size_type p = line.find('\r');
    if(p != line.npos) line.erase(p, 1);

    std::string output, bounce;

    ConvertMessage(line, output, bounce);
    if(!output.empty()) outf.Write(output);
    if(!bounce.empty()) inf.Write(bounce);

    return false;
}

static SockStream ConnectIRC_IPv4()
{
    struct sockaddr_in MyAddr;

    struct hostent *HostName = gethostbyname(GetServer().c_str());
    if(!HostName)
    {
        throw "ConnectIRC_IPv4::gethostbyname";
    }

    memset(&MyAddr, 0, sizeof(MyAddr));

    MyAddr.sin_family = AF_INET;
    MyAddr.sin_port = htons(GetServerPort());
    memcpy(&MyAddr.sin_addr, HostName->h_addr, HostName->h_length);

    int Plug = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(Plug < 0)
    {
        throw "ConnectIRC_IPv4::socket";
    }

    if(connect(Plug, (struct sockaddr*)&MyAddr, sizeof(MyAddr)) == -1)
    {
        close(Plug);
        throw "ConnectIRC_IPv4::connect";
    }

    return SockStream(Plug);
}

static SockStream ConnectIRC_IPv6()
{
    struct sockaddr *sa = NULL;

    if (!(sa = ResolveHost(GetServer().c_str(), AF_UNSPEC)))
    {
        throw "ConnectIRC_IPv6::ResolveHost";
    }

    ((struct sockaddr_in*) sa)->sin_port = htons(GetServerPort());

    int Plug = socket(sa->sa_family, SOCK_STREAM, IPPROTO_TCP);
    if(Plug < 0)
    {
        throw "ConnectIRC_IPv6::socket";
    }

    if(connect(Plug, sa, (sa->sa_family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6)) == -1)
    {
        close(Plug);
        throw "ConnectIRC_IPv6::connect";
    }

    return SockStream(Plug);
}

void ReportSocketError(const char* msg)
{
#ifndef WIN32
    perror(msg);
#else
    int err = errno;
    printf("%s: error %d ", msg, WSAGetLastError());
    printf("'%s'\n", sys_errlist[err]);
    //WSACleanUp();
#endif
}


#ifndef WIN32
/* POSIX version */
static void ClientLoop(SockStream& user, SockStream& server)
{
    for(;;)
    {
        struct pollfd polls[2] =
        {
            {user.GetFD(), POLLIN, 0},
            {server.GetFD(), POLLIN, 0}
        };
        int n = poll(polls, 2, -1);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(polls[0].revents & POLLIN)
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(polls[1].revents & POLLIN)
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
}
#else
/* Windows version */
static void ClientLoop(SockStream& user, SockStream& server)
{
    for(;;)
    {
        int u = user.GetFD();
        int s = server.GetFD();
        int upper = u>s ? u : s;

        fd_set r;
        FD_ZERO(&r);
        FD_SET(u, &r);
        FD_SET(s, &r);

        int n = select(upper+1, &r, NULL, NULL, NULL);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(FD_ISSET(u, &r))
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(FD_ISSET(s, &r))
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
}
#endif

static void Service(SockStream user)
{
    SockStream server;
    try
    {
        try
        {
            server = ConnectIRC_IPv6();
        }
        catch(const char *s)
        {
            user.Write("ERROR :" + std::string(s) + "\n");
            server = ConnectIRC_IPv4();
        }
    }
    catch(const char *s)
    {
        char Buf[512];

        std::string error = "Can not connect to ";
        error += GetServer();
        error += ':';
        sprintf(Buf, "%u", GetServerPort());
        error += Buf;

        error += ": ";
        error += s;
        error += ": ";
        if(errno != 0)
        {
            error += strerror(errno);
        }

        error = "ERROR :" + error + "\n";
        user.Write(error);
        return;
    }

    ClientLoop(user, server);
}

#ifndef WIN32
/* POSIX version */
static void CatchDead(int);
static void LaunchSocketHandler(int client, int listener)
{
    /* Listen to child deaths */
    signal(SIGCHLD, CatchDead);

    int forkpid = fork();
    switch(forkpid)
    {
        case -1:
        {
            /* Failed? */
            perror("fork");
            break;
        }
        case 0:
        {
            /* This starts a child process. */

            /* Child closes his unneeded resources. */
            close(listener);
            close(STDERR_FILENO);
            /* Child doesn't except child deaths. */
            signal(SIGCHLD, SIG_IGN);
            /* Child starts handling the client. */
            Service(client);
            /* Client has terminated. End. */
            _exit(0);
        }
    }
    /* In any case, the main process doesn't
     * need to worry about the client anymore.
     */
    close(client);
}
static void CatchDead(int)
{
    /* Wait for the dead. */
    waitpid(-1, NULL, WNOHANG);
}
#else
/* Windows version */
static DWORD WINAPI WindowsClientThread(LPVOID param)
{
    Service((int)param);
    return 0;
}
static void LaunchSocketHandler(int client, int listener)
{
    listener=listener; /* unused */
    DWORD threadid;
    CloseHandle(CreateThread
        (NULL, 0, WindowsClientThread,
         (LPVOID)(client),
         0,
         &threadid));
}
#endif


int main(int argc, char**argv)
{
#ifdef WIN32
    WSADATA wsaData;
    if(WSAStartup(MAKEWORD(2,2),&wsaData) == SOCKET_ERROR)
    {
        fprintf(stderr,"WSAStartup failed with error %d\n", WSAGetLastError());
        WSACleanup();
        return -1;
    }
#endif
    printf(
"iconvircproxy version "VERSION"\n"
"Copyright (C) 1992,2004 Joel Yliluoma (http://iki.fi/bisqwit/)\n"
"\n"
"iconvircproxy is free software, covered by the General Public License,\n"
"and you are welcome to change it and/or distribute copies of it under\n"
"certain conditions. Read the COPYING file to see the conditions.\n"
"There is absolutely no warranty for iconvircproxy.\n");

    /* Close everything but stderr */
#ifndef WIN32
    for(int fd=0;fd<512;++fd) { if(fd != STDERR_FILENO) close(fd); }
#endif
    struct sockaddr_in They;
    int Plug = -1;
    if(true) // Try IPv6
    {
        struct sockaddr_in6 MyAddr;

        memset(&MyAddr, 0, sizeof(MyAddr));

        MyAddr.sin6_family = AF_INET6;
        MyAddr.sin6_port = htons(GetListenPort());

        int Plug = socket(AF_INET6, SOCK_STREAM, IPPROTO_TCP);
        if(Plug < 0)
            ReportSocketError("main::socket");
        else
        {
        #ifndef WIN32
            /* Win32 doesn't support setsockopt? */
            int flag=1; /* Just a temp variable, boolean value for setsockopt() */
            setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof flag);
        #endif

            if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(MyAddr)))
            {
                ReportSocketError("main::bind");
                close(Plug);
                Plug = -1;
            }
        }
    }
    if(Plug == -1) // Try IPv4
    {
        struct sockaddr_in MyAddr;

        memset(&MyAddr, 0, sizeof(MyAddr));

        MyAddr.sin_family = AF_INET;
        MyAddr.sin_port = htons(GetListenPort());
        MyAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* auto-fill with my IP, network long  */

        int Plug = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if(Plug < 0) { ReportSocketError("main::socket"); return -errno; }

    #ifndef WIN32
        /* Win32 doesn't support setsockopt? */
        int flag=1; /* Just a temp variable, boolean value for setsockopt() */
        setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof flag);
    #endif
        if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)))
        {
            ReportSocketError("main::bind");
            close(Plug);
            return -errno;
        }
    }

    if(listen(Plug, 10))
    {
        close(Plug);
        ReportSocketError("main::listen");
        return -errno;
    }

    /* Flush error messages from undefined charsets. */
    FindChain2For("");

    argc=argc;
    printf("%s started successfully. Listening to connections.\n", argv[0]);
    for(;;)
    {
#ifndef WIN32
        unsigned AddrLen = sizeof(They);
#else
        int AddrLen = sizeof(They);
#endif
        int Sock = accept(Plug, (struct sockaddr*)&They, &AddrLen);
        if(Sock < 0)
        {
            if(errno == EINTR || errno == EAGAIN) continue;

            ReportSocketError("accept");
            break;
        }

        LaunchSocketHandler(Sock, Plug);
    }
    close(Plug);
    return 0;
}
